@@ -2,16 +2,18 @@ from datetime import datetime, timedelta
 from pypac import PACSession, get_pac
 import sys
 import re
+import logging
+from time import sleep
 
 
 class BaseScrape:
     def __init__(self,
-        use_proxy:bool=False,
-        proxy:str=None,
-        pac_url:str=None
-        ):
+                 use_proxy: bool = False,
+                 proxy: str = None,
+                 pac_url: str = None
+                 ):
         """Class for scrapping webpages
-        
+
         args:
         ----
         use_proxy - bool:False - should a proxy be used
@@ -27,7 +29,8 @@ class BaseScrape:
         self.time_marks = []
         self.use_proxy = use_proxy
         self.proxy = proxy
-        self.pac = get_pac(url=pac_url) if self.use_proxy else None
+        self.pac = get_pac(url=pac_url) \
+            if self.use_proxy and pac_url is not None else None
         self.pac_session = None
         # Variables for randomly generating headers
         self.header_vars = None
@@ -47,7 +50,7 @@ class BaseScrape:
 
     def _get_proxies(self, url):
         if self.use_proxy:
-            #use pypac
+            # use pypac
             proxies = self.pac_session \
                 ._get_proxy_resolver(self.pac) \
                 .get_proxy_for_requests(url)
@@ -58,34 +61,71 @@ class BaseScrape:
     def _get_proxy(self, url):
         if self.use_proxy:
             if self.proxy:
-                #use given proxy
+                # use given proxy
                 proxy = self.proxy
             elif self.pac:
-                #use pypac
+                # use pypac
                 proxies = self._get_proxies(url)
                 match = re.search(r"^(\w*)", str(url))
                 proxy = proxies[match.group()]
             else:
-                raise ValueError("Either pac_url or a proxy must being given in order for use_proxy to be True")
+                raise ValueError(
+                    "Either pac_url or a proxy must being given in order for use_proxy to be True")
         else:
             proxy = None
         return proxy
 
+    def handle_responses(self, scrape_urls, resps, scrape_resps, init_len):
+        # Add scrape_resps to resps
+        resps |= {r["url"]: r for r in scrape_resps}
+        # Get scraped urls
+        # Split success and fails
+        success_urls = set([
+            r["url"] for r in scrape_resps
+            if not r["error"]
+        ])
+        errored_urls = set([
+            r["url"] for r in scrape_resps
+            if r["error"]
+        ])
+        # Increment attempts count on each scraped url
+        self._increment_attempts(True, scrape_urls)
+        # Increment attempts count on each attempted but failed (IE had an error
+        # but not cancelled)
+        self._increment_attempts(False, errored_urls)
+        # Remove scraped urls from scrape_urls
+        scrape_urls = scrape_urls.difference(success_urls)
+        # Remove urls where too many attempts have been made
+        failed_urls = set(self.tracker_df[
+            self.tracker_df.attempts >= self.attempt_limit
+        ].url.to_list())
+        scrape_urls = scrape_urls.difference(failed_urls)
+        logging.info(f"""Scraping round complete, summary:
+    attempted:                               {init_len}
+    successful scrapes:                      {len(success_urls)}
+    errored scrapes (will attempt again):    {len(errored_urls)}
+    failed scrapes (will not attempt again): {len(failed_urls)}
+    remaining urls:                          {len(scrape_urls)}""")
+        return [scrape_urls, resps, failed_urls]
+
     def reset_pages_scraped(self):
         self.pages_scraped = 0
         self.time_marks = [datetime.now()]
 
     def increment_pages_scraped(self):
         self.pages_scraped += 1
-        #Add time mark
+        # Add time mark
         self.time_marks.append(datetime.now())
-        #Calc loop time
-        loop_time_elapsed = (self.time_marks[-1] - self.time_marks[-2]).total_seconds()
-        #Calc estimated finish time
+        # Calc loop time
+        loop_time_elapsed = (
+            self.time_marks[-1] - self.time_marks[-2]).total_seconds()
+        # Calc estimated finish time
         total_time_elapsed = (datetime.now() - self.job_start).total_seconds()
-        est_total_time_s = total_time_elapsed + (self.total_to_scrape * total_time_elapsed / self.pages_scraped)
-        #Output
-        sys.stdout.write(f"\rprocessed -> {self.pages_scraped}/{self.total_to_scrape} - loop time {loop_time_elapsed:.6f} - est finish {datetime.now() + timedelta(seconds=est_total_time_s)}\n")
+        est_total_time_s = total_time_elapsed + \
+            (self.total_to_scrape * total_time_elapsed / self.pages_scraped)
+        # Output
+        sys.stdout.write(
+            f"\rprocessed -> {self.pages_scraped}/{self.total_to_scrape} - loop time {loop_time_elapsed:.6f} - est finish {datetime.now() + timedelta(seconds=est_total_time_s)}\n")
         sys.stdout.flush()
 
     def throttle_tasks(self):
@@ -100,9 +140,9 @@ class BaseScrape:
     def end_job(self):
         self.job_end = datetime.now()
         runtime = self.job_end - self.job_start
-        print(f"Job completed in {runtime}")
+        logging.info(f"Job completed in {runtime}")
 
     def _get_pac_session(self):
         if not self.pac_session:
             self.pac_session = PACSession(self.pac)
-        return self.pac_session
\ No newline at end of file
+        return self.pac_session
