@@ -2,11 +2,12 @@ from time import sleep
 import pandas as pd
 from requests_html import HTMLSession
 import logging
-import sys
-
+from requests import Response
+from http.client import HTTPResponse
 
 from .base_scrape import BaseScrape
 from ..utils.header_vars import random_header_vars
+from ..utils.errors import HttpResponseStatusError
 
 
 class Scrape(BaseScrape):
@@ -58,7 +59,7 @@ class Scrape(BaseScrape):
         self.fetch_error_handler = fetch_error_handler
         self.session = HTMLSession()
         # Define allowed errors
-        self.acceptable_errors = ()
+        self.acceptable_errors = (HttpResponseStatusError,)
         self.consecutive_error_limit = consecutive_error_limit
         self.consecutive_error_count = 0
         # Define criteria for looping multiple attempts
@@ -73,6 +74,9 @@ class Scrape(BaseScrape):
         self._get_pac_session()
         self.session = self.pac_session
 
+    def _request(self, url: str):
+        return self.session.get(url, headers=self.headers)
+
     def _fetch(self, url):
         """Function to fetch HTML from url
 
@@ -88,16 +92,25 @@ class Scrape(BaseScrape):
         # Make the request
         try:
             if url:
-                resp = self.session.get(url, headers=self.headers)
-                if resp.status_code == 200:
+                resp = self._request(url)
+                if isinstance(resp, Response):
+                    status = resp.status_code
                     html = resp.text
-                    func_resp = self.post_process(
-                        html=html, resp=resp, **self.post_process_kwargs)
+                elif isinstance(resp, HTTPResponse):
+                    status = resp.status
+                    html = resp.read()
                 else:
-                    func_resp = None
+                    raise TypeError(
+                        "resp should be of type HTTPResponse or Response")
+                if status != 200:
+                    raise HttpResponseStatusError(
+                        f"url responded with a http status of {status}")
+                func_resp = self.post_process(
+                    html=html, resp=resp, **self.post_process_kwargs) \
+                    if html is not None else None
                 # Reset self.acceptable_error_count if all goes fine
                 self.consecutive_error_count = 0
-                return {"url": url, "func_resp": func_resp, "status": resp.status_code, "error": None}
+                return {"url": url, "func_resp": func_resp, "status": status, "error": None}
         except Exception as e:
             # Set the current error - increment if the same error
             if type(e) == self.cur_err:
@@ -166,13 +179,14 @@ class Scrape(BaseScrape):
             {"url": u, "scraped": False, "attempts": 0}
             for u in urls
         ])
-        resps = []
-        i = 0
-        scrape_urls = urls
+        resps = {}
+        scrape_urls = set(urls)
+        logging.info(f"{len(urls)} unique urls from {len(scrape_urls)}")
         while len(scrape_urls):
+            init_len = len(scrape_urls)
             self.reset_pages_scraped()
             self.total_to_scrape = len(scrape_urls)
-            # Runt he scrapes
+            # Run the scrapes
             scrape_resps = []
             for url in urls:
                 scrape_resps.append(self._fetch(url))
@@ -180,40 +194,10 @@ class Scrape(BaseScrape):
                 # Regenerate headers
                 if self.randomise_headers:
                     self.headers = random_header_vars(self.header_vars)
-            # Add scrape_resps to resps
-            resps.extend(scrape_resps)
-            # Get scraped urls
-            scraped_urls = [
-                r["url"] for r in resps
-                if not r["error"]
-            ]
-            # Increment attempts count on each scraped url
-            self._increment_attempts(True, scrape_urls)
-            # Get errored urls
-            errored_urls = [
-                r["url"] for r in resps
-                if r["error"]
-            ]
-            # Increment attempts count on each attempted but failed (IE had an error
-            # but not cancelled)
-            self._increment_attempts(False, errored_urls)
-            # Remove scraped urls from scrape_urls
-            scrape_urls = [
-                u for u in scrape_urls
-                if u not in scraped_urls
-            ]
-            # Remove urls where too many attempts have been made
-            failed_urls = self.tracker_df[
-                self.tracker_df.attempts >= self.attempt_limit
-            ].url.to_list()
-            scrape_urls = [
-                u for u in scrape_urls
-                if u not in failed_urls
-            ]
-            logging.info(
-                f"Scraping round {i} complete, in this round - {len(scraped_urls)} urls scrapped, {len(errored_urls)} urls errored, {len(errored_urls) - len(scrape_urls)} urls failed, {len(scrape_urls)} remain unscrapped")
-            # Increment the loop number
-            i += 1
+            # Process responses
+            scrape_urls, resps, failed_urls = \
+                self.handle_responses(scrape_urls, resps,
+                                      scrape_resps, init_len)
             # Sleep before running again
             # - shutdown must have been initiated
             # - there must still be urls to scrape
@@ -222,7 +206,10 @@ class Scrape(BaseScrape):
                     and self.rest_between_attempts:
                 logging.info(f"Sleeping for {self.rest_wait} seconds")
                 sleep(self.rest_wait)
-        logging.info(f"Scraping complete {len(failed_urls)} urls failed")
+        logging.info(
+            f"Scraping complete {len(failed_urls)}/{len(set(urls))} urls failed")
+        # Convert resps back
+        resps = [v for _, v in resps.items()]
         # end the job
         self.end_job()
         return resps
